#
# Solvers
#
import liionpack as lp
from liionpack.solver_utils import _create_casadi_objects as cco
from liionpack.solver_utils import _serial_step as ss
from liionpack.solver_utils import _mapped_step as ms
from liionpack.solver_utils import _serial_eval as se
from liionpack.solver_utils import _mapped_eval as me
import ray
import numpy as np
import time as ticker
from tqdm import tqdm
import pybamm


class GenericActor:
    def __init__(self):
        pass

    def setup(
        self,
        Nspm,
        sim_func,
        parameter_values,
        dt,
        inputs,
        variable_names,
        initial_soc,
        nproc,
    ):
        # Casadi specific arguments
        if nproc > 1:
            mapped = True
        else:
            mapped = False
        self.Nspm = Nspm
        # Set up simulation
        self.parameter_values = parameter_values
        if initial_soc is not None:
            if (
                (type(initial_soc) in [float, int])
                or (isinstance(initial_soc, list) and len(initial_soc) == 1)
                or (isinstance(initial_soc, np.ndarray) and len(initial_soc) == 1)
            ):
                _, _ = lp.update_init_conc(parameter_values, initial_soc, update=True)
            else:
                lp.logger.warning(
                    "Using a list or an array of initial_soc "
                    + "is not supported, please set the initial "
                    + "concentrations via inputs"
                )
        if sim_func is None:
            self.simulation = lp.basic_simulation(self.parameter_values)
        else:
            self.simulation = sim_func(self.parameter_values)

        # Set up integrator
        casadi_objs = cco(
            inputs, self.simulation, dt, Nspm, nproc, variable_names, mapped
        )
        self.model = self.simulation.built_model
        self.integrator = casadi_objs["integrator"]
        self.variables_fn = casadi_objs["variables_fn"]
        self.t_eval = casadi_objs["t_eval"]
        self.event_names = casadi_objs["event_names"]
        self.events_fn = casadi_objs["events_fn"]
        self.step_solutions = casadi_objs["initial_solutions"]
        self.last_events = None
        self.event_change = None
        if mapped:
            self.step_fn = ms
            self.eval_fn = me
        else:
            self.step_fn = ss
            self.eval_fn = se

    def step(self, inputs):
        # Solver Step
        self.step_solutions, self.var_eval, self.events_eval = self.step_fn(
            self.simulation.built_model,
            self.step_solutions,
            inputs,
            self.integrator,
            self.variables_fn,
            self.t_eval,
            self.events_fn,
        )
        return self.check_events()

    def evaluate(self, inputs):
        self.var_eval = self.eval_fn(
            self.simulation.built_model,
            self.step_solutions,
            inputs,
            self.variables_fn,
            self.t_eval,
        )

    def check_events(self):
        if self.last_events is not None:
            # Compare changes
            new_sign = np.sign(self.events_eval)
            old_sign = np.sign(self.last_events)
            self.event_change = (old_sign * new_sign) < 0
            self.last_events = self.events_eval
            return np.any(self.event_change)
        else:
            self.last_events = self.events_eval
            return False

    def get_event_change(self):
        return self.event_change

    def get_event_names(self):
        return self.event_names

    def output(self):
        return self.var_eval


@ray.remote(num_cpus=1)
class RayActor(GenericActor):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class GenericManager:
    def __init__(
        self,
    ):
        pass

    def solve(
        self,
        netlist,
        sim_func,
        parameter_values,
        experiment,
        inputs,
        output_variables,
        initial_soc,
        nproc,
        node_termination_func=None,
        setup_only=False,
    ):
        self.netlist = netlist
        self.sim_func = sim_func
        self.node_termination_func = node_termination_func
        self.parameter_values = parameter_values
        self.check_current_function()
        # Get netlist indices for resistors, voltage sources, current sources
        self.Ri_map = netlist["desc"].str.find("Ri") > -1
        self.V_map = netlist["desc"].str.find("V") > -1
        self.I_map = netlist["desc"].str.find("I") > -1
        self.Terminal_Node = np.array(netlist[self.I_map].node1)
        self.Nspm = np.sum(self.V_map)

        self.split_models(self.Nspm, nproc)

        # Generate the protocol from the supplied experiment
        self.protocol_steps, self.terminations, self.step_types = (
            lp.generate_protocol_from_experiment(experiment)
        )
        self.flattened_protocol = [
            item for sublist in self.protocol_steps for item in sublist
        ]
        self.dt = experiment.period
        self.Nsteps = len(self.flattened_protocol)
        # If the step is starting with a rest the current will be zero and
        # this messes up the internal resistance calc. Add a very small current
        # for init.
        first_value = self.protocol_steps[0][0]
        if first_value == 0.0:
            netlist.loc[self.I_map, ("value")] = 1e-6
        else:
            netlist.loc[self.I_map, ("value")] = first_value
        # Solve the circuit to initialise the electrochemical models
        if self.step_types[0] == "power":
            current = None
            power = first_value
        else:
            current = first_value
            power = None

        V_node, I_batt, terminal_current, terminal_voltage, terminal_power = (
            lp.solve_circuit(netlist, current=current, power=power)
        )

        # The simulation output variables calculated at each step for each battery
        # Must be a 0D variable i.e. battery wide volume average - or X-averaged for
        # 1D model
        self.variable_names = [
            "Terminal voltage [V]",
            "Surface open-circuit voltage [V]",
        ]
        if output_variables is not None:
            for out in output_variables:
                if out not in self.variable_names:
                    self.variable_names.append(out)
            # variable_names = variable_names + output_variables
        self.Nvar = len(self.variable_names)

        # Storage variables for simulation data
        self.shm_i_app = np.zeros([self.Nsteps, self.Nspm], dtype=np.float32)
        self.shm_Ri = np.zeros([self.Nsteps, self.Nspm], dtype=np.float32)
        self.output = np.zeros([self.Nvar, self.Nsteps, self.Nspm], dtype=np.float32)
        self.node_voltages = np.zeros([self.Nsteps, len(V_node)], dtype=np.float32)

        # Initialize currents in battery models
        self.shm_i_app[0, :] = I_batt * -1

        # Initialize the node voltages
        self.node_voltages[0, :] = V_node

        # Step forward in time
        self.V_terminal = np.zeros(self.Nsteps, dtype=np.float32)
        self.I_terminal = np.zeros(self.Nsteps, dtype=np.float32)
        self.P_terminal = np.zeros(self.Nsteps, dtype=np.float32)
        self.record_times = np.zeros(self.Nsteps, dtype=np.float32)

        self.v_cut_lower = parameter_values["Lower voltage cut-off [V]"]
        self.v_cut_higher = parameter_values["Upper voltage cut-off [V]"]

        # Handle the inputs
        self.inputs = inputs
        self.inputs_dict = lp.build_inputs_dict(self.shm_i_app[0, :], self.inputs, None)
        # Solver specific setup
        self.setup_actors(nproc, self.inputs_dict, initial_soc)
        # Get the initial state of the system
        self.evaluate_actors()
        if not setup_only:
            self.global_step = 0
            for ps, step_protocol in enumerate(self.protocol_steps):
                step_termination = self.terminations[ps]
                step_type = self.step_types[ps]
                if step_termination == []:
                    step_termination = 0.0
                self._step_solve_step(step_protocol, step_termination, step_type, None)
            return self.step_output()

    def _step_solve_step(self, protocol, termination, step_type, updated_inputs):
        tic = ticker.time()

        # Do stepping
        lp.logger.notice("Starting step solve")
        vlims_ok = True
        skip_vcheck = True
        with tqdm(total=len(protocol), desc="Stepping simulation") as pbar:
            step = 0
            while step < len(protocol):
                vlims_ok = self._step(
                    step, protocol, termination, step_type, updated_inputs, skip_vcheck
                )
                skip_vcheck = False
                if vlims_ok:
                    # all good - keep going
                    self.global_step += 1
                    step += 1
                    pbar.update(1)
                else:
                    # Move on to next protocol step
                    break
        self.step = step
        toc = ticker.time()
        lp.logger.notice("Step solve finished")
        lp.logger.notice("Total stepping time " + str(np.around(toc - tic, 3)) + "s")
        lp.logger.notice(
            "Time per step " + str(np.around((toc - tic) / len(protocol), 3)) + "s"
        )

    def step_output(self):
        self.cleanup()
        self.shm_Ri = np.abs(self.shm_Ri)
        # Collect outputs
        report_steps = min(len(self.flattened_protocol), self.global_step)
        self.all_output = {}
        self.all_output["Time [s]"] = self.record_times[:report_steps]
        self.all_output["Pack current [A]"] = self.I_terminal[:report_steps]
        self.all_output["Pack terminal voltage [V]"] = self.V_terminal[:report_steps]
        self.all_output["Pack power [W]"] = self.P_terminal[:report_steps]
        self.all_output["Cell current [A]"] = self.shm_i_app[:report_steps, :]
        self.all_output["Node voltage [V]"] = self.node_voltages[:report_steps, :]
        self.all_output["Cell internal resistance [Ohm]"] = self.shm_Ri[
            :report_steps, :
        ]
        for j in range(self.Nvar):
            self.all_output[self.variable_names[j]] = self.output[j, :report_steps, :]
        return self.all_output

    def _pack_voltage(self, step):
        current_nodes = self.netlist.loc[
            self.I_map, (["node2", "node1"])
        ].values.flatten()
        return np.diff(self.node_voltages[step, current_nodes])[0]

    def _step(
        self, step, protocol, termination, step_type, updated_inputs, skip_vcheck
    ):
        vlims_ok = True
        # 01 Calculate whether resting or restarting
        self.resting = step > 0 and protocol[step] == 0.0 and protocol[step - 1] == 0.0
        self.restarting = (
            step > 0 and protocol[step] != 0.0 and protocol[step - 1] == 0.0
        )
        # 02 Get the actor output - Battery state info
        self.get_actor_output(self.global_step)
        # 03 Get the ocv and internal resistance
        temp_v = self.output[0, self.global_step, :]
        temp_ocv = self.output[1, self.global_step, :]
        # When resting and rebalancing currents are small the internal
        # resistance calculation can diverge as it's R = V / I
        # At rest the internal resistance should not change greatly
        # so for now just don't recalculate it.
        if not self.resting and not self.restarting:
            self.temp_Ri = self.calculate_internal_resistance(self.global_step)
        self.shm_Ri[self.global_step, :] = self.temp_Ri
        # 04 Update netlist
        self.netlist.loc[self.V_map, ("value")] = temp_ocv
        self.netlist.loc[self.Ri_map, ("value")] = self.temp_Ri

        # 05 Solve the circuit with updated netlist
        if step <= self.Nsteps:
            if step_type == "power":
                power = protocol[step]
                current = None
            else:
                current = protocol[step]
                power = None
            V_node, I_batt, terminal_current, terminal_voltage, terminal_power = (
                lp.solve_circuit(self.netlist, current=current, power=power)
            )
            lp.power_loss(self.netlist)
            self.record_times[self.global_step] = self.global_step * self.dt
            self.netlist.loc[self.I_map, ("value")] = terminal_current
            self.I_terminal[self.global_step] = terminal_current[0]
            self.V_terminal[self.global_step] = terminal_voltage[0]
            self.P_terminal[self.global_step] = terminal_power[0]
        if self.global_step < self.Nsteps - 1:
            # igore last step save the new currents and build inputs
            # for the next step
            I_app = I_batt[:] * -1
            self.shm_i_app[self.global_step, :] = I_app
            self.shm_i_app[self.global_step + 1, :] = I_app
            self.node_voltages[self.global_step, :] = V_node
            self.inputs_dict = lp.build_inputs_dict(I_app, self.inputs, updated_inputs)
        # 06 Check if voltage limits are reached and terminate
        if np.any(temp_v < self.v_cut_lower):
            lp.logger.warning("Low voltage limit reached")
            vlims_ok = False
        if np.any(temp_v > self.v_cut_higher):
            lp.logger.warning("High voltage limit reached")
            vlims_ok = False
        v_thresh = temp_v - termination
        if np.any(v_thresh < 0) and np.any(v_thresh > 0):
            # some have crossed the stopping condition
            vlims_ok = False
        if self.node_termination_func is not None:
            if self.node_termination_func(V_node):
                lp.logger.warning("Node voltage limit reached")
                vlims_ok = False
        if skip_vcheck:
            vlims_ok = True
        # 07 Step the electrochemical system
        self.step_actors()
        return vlims_ok

    def check_current_function(self):
        i_func = self.parameter_values["Current function [A]"]
        if i_func.__class__ is not pybamm.InputParameter:
            self.parameter_values.update({"Current function [A]": "[input]"})
            lp.logger.notice(
                "Parameter: Current function [A] has been set to " + "input"
            )

    def actor_i_app(self, index):
        actor_indices = self.split_index[index]
        return self.shm_i_app[self.timestep, actor_indices]

    def actor_htc(self, index):
        return self.htc[index]

    def build_inputs(self):
        inputs = []
        for i in range(len(self.actors)):
            inputs.append(self.inputs_dict[self.slices[i]])
        return inputs

    def calculate_internal_resistance(self, step):
        # Calculate internal resistance and update netlist
        temp_v = self.output[0, step, :]
        temp_ocv = self.output[1, step, :]
        temp_I = self.shm_i_app[step, :]
        temp_Ri = np.abs((temp_ocv - temp_v) / temp_I)
        temp_Ri[temp_Ri == 0.0] = 1e-6
        return temp_Ri

    def update_external_variables(self):
        # This is probably going to involve reading from disc unless the whole
        # algorithm is wrapped inside an "external" solver
        # For now use a dummy function to test changing the values
        pass

    def split_models(self, Nspm, nproc):
        pass

    def setup_actors(self, nproc, inputs, initial_soc):
        pass

    def step_actors(self):
        pass

    def evaluate_actors(self):
        pass

    def get_actor_output(self, step):
        pass

    def cleanup(self):
        pass


class RayManager(GenericManager):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        lp.logger.notice("Ray initialization started")
        ray.init()
        lp.logger.notice("Ray initialization complete")

    def split_models(self, Nspm, nproc):
        # Manage the number of SPM models per worker
        self.split_index = np.array_split(np.arange(Nspm), nproc)
        self.spm_per_worker = [len(s) for s in self.split_index]
        self.slices = []
        for i in range(nproc):
            self.slices.append(
                slice(self.split_index[i][0], self.split_index[i][-1] + 1)
            )

    def setup_actors(self, nproc, inputs, initial_soc):
        tic = ticker.time()
        # Ray setup an actor for each worker
        self.actors = []
        for i in range(nproc):
            self.actors.append(lp.RayActor.remote())
        setup_futures = []
        for i, a in enumerate(self.actors):
            # Create actor on each worker containing a simulation
            setup_futures.append(
                a.setup.remote(
                    Nspm=self.spm_per_worker[i],
                    sim_func=self.sim_func,
                    parameter_values=self.parameter_values,
                    dt=self.dt,
                    inputs=inputs[self.slices[i]],
                    variable_names=self.variable_names,
                    initial_soc=initial_soc,
                    nproc=1,
                )
            )
        _ = [ray.get(f) for f in setup_futures]
        toc = ticker.time()
        lp.logger.notice(
            "Ray actors setup in time " + str(np.around(toc - tic, 3)) + "s"
        )

    def step_actors(self):
        t1 = ticker.time()
        future_steps = []
        inputs = self.build_inputs()
        for i, pa in enumerate(self.actors):
            future_steps.append(pa.step.remote(inputs[i]))
        events = [ray.get(fs) for fs in future_steps]
        if np.any(events):
            self.log_event()
        t2 = ticker.time()
        lp.logger.info("Ray actors stepped in " + str(np.around(t2 - t1, 3)) + "s")

    def evaluate_actors(self):
        t1 = ticker.time()
        future_evals = []
        inputs = self.build_inputs()
        for i, pa in enumerate(self.actors):
            future_evals.append(pa.evaluate.remote(inputs[i]))
        _ = [ray.get(fs) for fs in future_evals]
        t2 = ticker.time()
        lp.logger.info("Ray actors evaluated in " + str(np.around(t2 - t1, 3)) + "s")

    def get_actor_output(self, step):
        t1 = ticker.time()
        futures = []
        for actor in self.actors:
            futures.append(actor.output.remote())
        for i, f in enumerate(futures):
            out = ray.get(f)
            self.output[:, step, self.split_index[i]] = out
        t2 = ticker.time()
        lp.logger.info(
            "Ray actor output retrieved in " + str(np.around(t2 - t1, 3)) + "s"
        )

    def log_event(self):
        futures = []
        for actor in self.actors:
            futures.append(actor.get_event_change.remote())
        all_event_changes = []
        for i, f in enumerate(futures):
            all_event_changes.append(np.asarray(ray.get(f)))
        event_change = np.hstack(all_event_changes)
        Nr, Nc = event_change.shape
        event_names = ray.get(self.actors[0].get_event_names.remote())
        for r in range(Nr):
            if np.any(event_change[r, :]):
                lp.logger.warning(
                    event_names[r]
                    + ", Batteries: "
                    + str(np.where(event_change[r, :])[0].tolist())
                )

    def cleanup(self):
        for actor in self.actors:
            ray.kill(actor)
        lp.logger.notice("Shutting down Ray")
        ray.shutdown()


class CasadiManager(GenericManager):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def split_models(self, Nspm, nproc):
        # For casadi there is no need to split the models as we pass them all
        # to the integrator however we still want the global variables to be
        # used in the same generic way
        self.spm_per_worker = Nspm
        self.split_index = np.array_split(np.arange(Nspm), 1)
        self.slices = [slice(self.split_index[0][0], self.split_index[0][-1] + 1)]

    def setup_actors(self, nproc, inputs, initial_soc):
        # For casadi we do not use multiple actors but instead the integrator
        # function that is generated by casadi handles multithreading behind
        # the scenes
        tic = ticker.time()

        self.actors = [GenericActor()]
        for a in self.actors:
            a.setup(
                Nspm=self.spm_per_worker,
                sim_func=self.sim_func,
                parameter_values=self.parameter_values,
                dt=self.dt,
                inputs=inputs,
                variable_names=self.variable_names,
                initial_soc=initial_soc,
                nproc=nproc,
            )
        toc = ticker.time()
        lp.logger.info(
            "Casadi actor setup in time " + str(np.around(toc - tic, 3)) + "s"
        )

    def step_actors(self):
        tic = ticker.time()
        events = self.actors[0].step(self.build_inputs()[0])
        if events:
            self.log_event()
        toc = ticker.time()
        lp.logger.info(
            "Casadi actor stepped in time " + str(np.around(toc - tic, 3)) + "s"
        )

    def evaluate_actors(self):
        tic = ticker.time()
        self.actors[0].evaluate(self.build_inputs()[0])
        toc = ticker.time()
        lp.logger.info(
            "Casadi actor evaluated in time " + str(np.around(toc - tic, 3)) + "s"
        )

    def get_actor_output(self, step):
        tic = ticker.time()
        self.output[:, step, :] = self.actors[0].output()
        toc = ticker.time()
        lp.logger.info(
            "Casadi actor output got in time " + str(np.around(toc - tic, 3)) + "s"
        )

    def log_event(self):
        event_change = np.asarray(self.actors[0].get_event_change())
        Nr, Nc = event_change.shape
        event_names = self.actors[0].get_event_names()
        for r in range(Nr):
            if np.any(event_change[r, :]):
                lp.logger.warning(
                    event_names[r]
                    + ", Batteries: "
                    + str(np.where(event_change[r, :])[0].tolist())
                )

    def cleanup(self):
        pass



import liionpack as lp
import pybamm
import numpy as np
import os
import matplotlib.pyplot as plt

plt.close("all")
lp.set_logging_level("NOTICE")

names = ["8P-1S", "4P-2S", "2S-4P", "2P-4S", "4S-2P", "8S-1P"]
case = 0
C = 5.0
# Define parameters

_Np = [8, 4, 4, 2, 2, 1]
_Ns = [1, 2, 2, 4, 4, 8]
_I = C * np.array(_Np)
configurations = [
    "parallel-strings",
    "parallel-strings",
    "series-groups",
    "parallel-strings",
    "series-groups",
    "parallel-strings",
]

Iapp = _I[case]
Np = _Np[case]
Ns = _Ns[case]
Nb = Np * Ns
configuration = configurations[case]
name = names[case]

# Generate the netlist
netlist = lp.setup_circuit(Np=Np, Ns=Ns, Rb=1e-4, I=C, configuration=configuration)
# lp.draw_circuit(netlist)

# Define additional output variables
output_variables = ["Volume-averaged cell temperature [K]"]

# Define a cycling experiment using PyBaMM
experiment = pybamm.Experiment(
    [
        f"Discharge at {Iapp} A for 1 hour or until 3.0 V",
        "Rest for 30 minutes",
        f"Charge at {Iapp / 2} A for 2 hours or until 4.1 V",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# Define the PyBaMM parameters
parameter_values = pybamm.ParameterValues("Chen2020")
parameter_values.update(
    {
        "Negative electrode active material volume fraction": "[input]",
        "Positive electrode active material volume fraction": "[input]",
    }
)
inputs = {
    "Total heat transfer coefficient [W.m-2.K-1]": 10 + (np.random.random(Nb) * 1.0),
    "Negative electrode active material volume fraction": 0.75
    + (np.random.random(Nb) * 0.01),
    "Positive electrode active material volume fraction": 0.66
    + (np.random.random(Nb) * 0.01),
}

# Solve the pack
output = lp.solve(
    netlist=netlist,
    sim_func=lp.thermal_simulation,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=None,
    inputs=inputs,
    nproc=os.cpu_count(),
    manager="casadi",
)

# Plot the pack and individual cell results
lp.plot_pack(output)
lp.plot_cells(output)

lp.show_plots()



#
# Test Power Controlled Circuit Solve
#

import liionpack as lp
import numpy as np
import matplotlib.pyplot as plt
import pybamm

plt.close("all")


# Define parameters
Np = 1
Ns = 3

# Generate netlist
netlist = lp.setup_circuit(Np=Np, Ns=Ns, Rb=1.5e-4, Rc=2e-2, Ri=5e-2, V=4.0, I=2.2)

current = 2.0
power = 10.0


I_map = netlist["desc"].str.find("I") > -1
Terminal_Node = np.array(netlist[I_map].node1)[0]

V_node, I_batt, terminal_current, terminal_voltage, terminal_power = lp.solve_circuit(
    netlist, current=10.0, power=None
)
V_Terminal = V_node[Terminal_Node]
print("Terminal voltage [V]:", terminal_voltage)
print("Current [A]:", terminal_current)
print("Power [W]:", terminal_power)

p = np.random.randint(5.0, 15.0, 100)
t = np.arange(len(p))

power_step = pybamm.experiment.step.power(50.0, duration="1000 s", period=10)
random_power_step = pybamm.experiment.step.power(np.vstack((t, p)).T)

# Cycling experiment
experiment = pybamm.Experiment(
    [power_step],
    period="1 s",
)
parameter_values = pybamm.ParameterValues("Chen2020")
output_variables = ["Terminal voltage [V]", "Current [A]", "Power [W]"]
# Solve pack
output = lp.solve(
    netlist=netlist,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=0.5,
)
lp.plot_output(output)



#
# Tests the logger class.
#
import liionpack as lp
import unittest
import os


class TestLogger(unittest.TestCase):
    def test_logger(self):
        logger = lp.logger
        self.assertEqual(logger.level, 30)
        lp.set_logging_level("INFO")
        self.assertEqual(logger.level, 20)
        lp.set_logging_level("ERROR")
        self.assertEqual(logger.level, 40)
        lp.set_logging_level("VERBOSE")
        self.assertEqual(logger.level, 15)
        lp.set_logging_level("NOTICE")
        self.assertEqual(logger.level, 25)
        lp.set_logging_level("SUCCESS")
        self.assertEqual(logger.level, 35)

        lp.set_logging_level("SPAM")
        self.assertEqual(logger.level, 5)
        lp.logger.spam("Test spam level")
        lp.logger.verbose("Test verbose level")
        lp.logger.notice("Test notice level")
        lp.logger.success("Test success level")

        # reset
        lp.set_logging_level("WARNING")

    def test_log_to_file(self):
        cwd = os.getcwd()
        filename = os.path.join(cwd, "temp_log_file.log")
        with open(filename, "w"):
            lp.log_to_file(filename)
            lp.logger.warning("This should write to file")
            assert os.path.isfile(filename)
        os.remove(filename)


if __name__ == "__main__":
    unittest.main()



"""
Example of using US06 drive cycle data for a battery pack simulation.
"""

import pybamm
import os
import pandas as pd
import liionpack as lp

os.chdir(pybamm.__path__[0] + "/..")

# Define parameters
Np = 4
Ns = 1

# Generate netlist
netlist = lp.setup_circuit(Np=Np, Ns=Ns, Rb=1.5e-3, Rc=1e-2, Ri=5e-2, V=4.0, I=5.0)

# Define the PyBaMM parameters
parameter_values = pybamm.ParameterValues("Chen2020")

# Import drive cycle from file
drive_cycle = pd.read_csv(
    pybamm.DataLoader().get_data("US06.csv"), comment="#", header=None
).to_numpy()

experiment = pybamm.Experiment(
    [pybamm.step.current(drive_cycle)],
    period="1 second",
)

# Solve pack
output = lp.solve(
    netlist=netlist,
    parameter_values=parameter_values,
    experiment=experiment,
    initial_soc=0.5,
)

# Plot results
lp.plot_output(output)
lp.show_plots()



import liionpack as lp
import numpy as np
import matplotlib.pyplot as plt
import unittest
import os


class netlist_utilsTest(unittest.TestCase):
    def test_read_netlist(self):
        net1 = lp.read_netlist("4p1s", I=50.0)
        net2 = lp.read_netlist("4p1s.txt", I=50.0)
        net3 = lp.read_netlist("4p1s.cir", I=50.0)
        I_map = net1["desc"].str.find("I") > -1
        assert np.all(net1[I_map]["value"] == 50.0)
        assert np.all(net2[I_map]["value"] == 50.0)
        assert np.all(net3[I_map]["value"] == 50.0)

    def test_netlist_exception(self):
        def bad_filename():
            _ = lp.read_netlist("4p1s.bad", I=50.0)

        with self.assertRaises(FileNotFoundError):
            bad_filename()

    def test_setup_circuit(self):
        netlist = lp.setup_circuit(Np=1, Ns=2, Rb=1e-4, Rc=1e-2, Ri=1e-3, V=2.0, I=10.0)
        V_map = netlist["desc"].str.find("V") > -1
        assert np.all(netlist[V_map]["value"] == 2)

    def test_circuit_configuration(self):
        netlist = lp.setup_circuit(
            Np=2,
            Ns=3,
            Rb=1e-4,
            Rc=1e-2,
            Ri=1e-3,
            V=2.0,
            I=10.0,
            configuration="parallel-strings",
        )
        assert sum(netlist["desc"].str.find("Rb") > -1) == 2
        netlist = lp.setup_circuit(
            Np=2,
            Ns=3,
            Rb=1e-4,
            Rc=1e-2,
            Ri=1e-3,
            V=2.0,
            I=10.0,
            configuration="series-groups",
        )
        assert sum(netlist["desc"].str.find("Rb") > -1) == 4

        def bad_configuration():
            _ = lp.setup_circuit(Np=2, Ns=1, Rb=1e-4, configuration="bad")

        with self.assertRaises(ValueError):
            bad_configuration()

    def test_setup_circuit_plot(self):
        netlist = lp.setup_circuit(
            Np=1, Ns=2, Rb=1e-4, Rc=1e-2, Ri=1e-3, V=2.0, I=10.0, plot=True
        )
        V_map = netlist["desc"].str.find("V") > -1
        assert np.all(netlist[V_map]["value"] == 2)
        plt.close("all")

    def test_solve_circuit_seriesgroups(self):
        netlist = lp.setup_circuit(
            Np=1,
            Ns=2,
            Rb=1e-4,
            Rc=1e-2,
            Ri=1e-3,
            V=2.0,
            I=1.0,
            configuration="series-groups",
        )
        V_node, I_batt, t_c, t_v, t_p = lp.solve_circuit(netlist)
        assert np.all(I_batt) == 1.0

    def test_setup_circuit_terminals(self):
        combos = [
            ["left", "right", "left-right", "right-left"],
            [[0, 0], [-1, -1], [0, -1], [-1, 0]],
        ]
        expected = [-1, 7, -1, -1]
        for terminals in combos:
            for i, t in enumerate(terminals):
                netlist = lp.setup_circuit(Np=7, Ns=1, Rb=1e-4, terminals=t)
                I_src = netlist[netlist["desc"] == "I0"]
                assert I_src["node1_x"].item() == expected[i]
                assert I_src["node2_x"].item() == expected[i]

        netlist = lp.setup_circuit(Np=7, Ns=1, Rb=1e-4, terminals=[4, 2])
        I_src = netlist[netlist["desc"] == "I0"]
        assert I_src["node1_x"].item() == -1
        assert I_src["node2_x"].item() == -1

    def test_terminals_exception(self):
        def bad_terminals():
            _ = lp.setup_circuit(Np=7, Ns=1, Rb=1e-4, terminals="bad")

        with self.assertRaises(ValueError):
            bad_terminals()

    def test_lcapy_circuit(self):
        l = lp.setup_circuit(Np=3, Ns=1, terminals="left")
        r = lp.setup_circuit(Np=3, Ns=1, terminals="right")
        lr = lp.setup_circuit(Np=3, Ns=1, terminals="left-right")
        rl = lp.setup_circuit(Np=3, Ns=1, terminals="right-left")
        m = lp.setup_circuit(Np=3, Ns=1, terminals=[1, 1])
        cct_l = lp.make_lcapy_circuit(l)
        cct_r = lp.make_lcapy_circuit(r)
        cct_lr = lp.make_lcapy_circuit(lr)
        cct_rl = lp.make_lcapy_circuit(rl)
        cct_m = lp.make_lcapy_circuit(m)
        assert cct_l.has_dc
        assert cct_r.has_dc
        assert cct_lr.has_dc
        assert cct_rl.has_dc
        assert cct_m.has_dc

    def test_write_netlist(self):
        net = lp.setup_circuit(Np=1, Ns=2, Rb=1e-4, Rc=1e-2, Ri=1e-3, V=2.0, I=10.0)
        cwd = os.getcwd()
        temp = os.path.join(cwd, "temp.txt")
        lp.write_netlist(net, temp)
        assert os.path.isfile(temp)
        os.remove(temp)


if __name__ == "__main__":
    unittest.main()



"""
Example of saving the output from a 16p2s battery pack simulation. Examples
are given for CSV, NumPy `.npy`, and NumPy `.npz` file formats.
"""

import liionpack as lp
import pybamm
import numpy as np
import os

# Define parameters
Np = 16
Ns = 2
Iapp = 20

# Generate the netlist
netlist = lp.setup_circuit(Np=Np, Ns=Ns)

# Define additional output variables
output_variables = ["Volume-averaged cell temperature [K]"]

# Define a cycling experiment using PyBaMM
experiment = pybamm.Experiment(
    [
        f"Charge at {Iapp} A for 30 minutes",
        "Rest for 15 minutes",
        f"Discharge at {Iapp} A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# Define the PyBaMM parameters
chemistry = pybamm.parameter_sets.Chen2020
parameter_values = pybamm.ParameterValues(chemistry=chemistry)
inputs = {"Total heat transfer coefficient [W.m-2.K-1]": np.ones(Np * Ns) * 10}

# Solve the pack
output = lp.solve(
    netlist=netlist,
    sim_func=lp.thermal_simulation,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=0.5,
    inputs=inputs,
    nproc=os.cpu_count(),
    manager="casadi",
)

# Save simulation output to CSV files
lp.save_to_csv(output)

# Save simulation output to Numpy npy files
lp.save_to_npy(output)

# Save simulation output to a compressed NumPy npz file
lp.save_to_npzcomp(output)



#
# A basic example of a pack simulation with varying external temperature.
#

import liionpack as lp
import pybamm
import numpy as np
import time as ticker
from tqdm import tqdm


lp.set_logging_level("NOTICE")

# Define parameters
Np = 16
Ns = 2
Nspm = Np * Ns
Iapp = 20

# Generate the netlist
netlist = lp.setup_circuit(Np=Np, Ns=Ns)

# Define additional output variables
output_variables = ["Volume-averaged cell temperature [K]"]

# Define a cycling experiment using PyBaMM
experiment = pybamm.Experiment(
    [
        f"Charge at {Iapp} A for 30 minutes",
        "Rest for 15 minutes",
        f"Discharge at {Iapp} A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# Define the PyBaMM parameters
parameter_values = pybamm.ParameterValues("Chen2020")
T0 = parameter_values["Initial temperature [K]"]

inputs = {}
input_temperature = np.ones(Nspm) * T0
inputs.update({"Input temperature [K]": input_temperature})
# Solve the pack

rm = lp.CasadiManager()
rm.solve(
    netlist=netlist,
    sim_func=lp.thermal_external,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    inputs=inputs,
    nproc=2,
    initial_soc=0.5,
    setup_only=True,
)


def external_stepper(manager, T0):
    tic = ticker.time()
    # Do stepping
    lp.logger.notice("Starting step solve")
    vlims_ok = True
    input_temperature = np.ones(Nspm) * T0
    with tqdm(total=manager.Nsteps, desc="Stepping simulation") as pbar:
        step = 0
        while step < manager.Nsteps and vlims_ok:
            input_temperature += 0.1
            inputs.update({"Input temperature [K]": input_temperature})
            vlims_ok = manager._step(step, inputs)
            step += 1
            pbar.update(1)
    manager.step = step
    toc = ticker.time()
    lp.logger.notice("Step solve finished")
    lp.logger.notice("Total stepping time " + str(np.around(toc - tic, 3)) + "s")
    lp.logger.notice(
        "Time per step " + str(np.around((toc - tic) / manager.Nsteps, 3)) + "s"
    )


external_stepper(rm, T0)
output = rm.step_output()
# Plot the pack and individual cell results
lp.plot_pack(output)
lp.plot_cells(output)
lp.show_plots()



"""
Simulate a big circuit containing 384 battery cells.
"""

import liionpack as lp
import pybamm
import os
import pickle

lp.log_to_file("logger_info")


if __name__ == "__main__":
    Np = 32
    Ns = 12
    Nspm = Np * Ns
    I_app = 5

    # Generate the netlist
    netlist = lp.setup_circuit(Np=Np, Ns=Ns, I=I_app)

    # Cycling experiment
    experiment = pybamm.Experiment(
        [
            f"Discharge at {I_app} A for 500 seconds",
            "Rest for 100 seconds",
            f"Charge at {I_app} A for 500 seconds",
            "Rest for 100 seconds",
        ],
        period="10 seconds",
    )

    # PyBaMM parameters
    parameter_values = pybamm.ParameterValues("Chen2020")

    # Solve pack
    output = lp.solve(
        netlist=netlist,
        parameter_values=parameter_values,
        experiment=experiment,
        nproc=os.cpu_count(),
    )

    with open("output.pickle", "wb") as handle:
        pickle.dump(output, handle, protocol=pickle.HIGHEST_PROTOCOL)



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.

import liionpack as lp
import pybamm


class BasicBenchmark:
    def setup(self):
        self.sim = lp.basic_simulation()

    def time_solve_model(self):
        self.sim.solve([0, 1800])


class SmallPack:
    timeout = 60

    def setup(self):
        self.netlist = lp.setup_circuit(Np=2, Ns=1, Rb=1e-4, Rc=1e-2)
        self.parameter_values = pybamm.ParameterValues("Chen2020")
        self.experiment = pybamm.Experiment(
            [
                "Discharge at 2 A for 5 minutes",
            ],
            period="10 seconds",
        )

    def time_discharge_1cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.experiment,
            initial_soc=0.5,
            nproc=1,
        )

    def time_discharge_2cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.experiment,
            initial_soc=0.5,
            nproc=2,
        )


class MediumPack:
    timeout = 120

    def setup(self):
        self.netlist = lp.setup_circuit(Np=32, Ns=10, Rb=1e-4, Rc=1e-2)
        self.parameter_values = pybamm.ParameterValues("Chen2020")
        self.experiment = pybamm.Experiment(
            [
                "Discharge at 32 A for 5 minutes",
            ],
            period="10 seconds",
        )

    def time_discharge_1cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.experiment,
            initial_soc=0.5,
            nproc=1,
        )

    def time_discharge_2cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.experiment,
            initial_soc=0.5,
            nproc=2,
        )


class LargePack:
    timeout = 600

    def setup(self):
        self.netlist = lp.setup_circuit(Np=64, Ns=10, Rb=1e-4, Rc=1e-2)
        self.parameter_values = pybamm.ParameterValues("Chen2020")
        self.experiment = pybamm.Experiment(
            [
                "Discharge at 64 A for 5 minutes",
            ],
            period="10 seconds",
        )
        I_app = 64
        self.long_experiment = pybamm.Experiment(
            [
                f"Charge at {I_app} A for 20 minutes",
                "Rest for 15 minutes",
                f"Discharge at {I_app} A for 20 minutes",
                "Rest for 30 minutes",
            ]
            * 3,
            period="10 seconds",
        )

    def time_discharge_1cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.experiment,
            initial_soc=0.5,
            nproc=1,
        )

    def time_discharge_2cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.experiment,
            initial_soc=0.5,
            nproc=2,
        )

    def time_long_cycle_2cpu(self):
        _ = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values.copy(),
            experiment=self.long_experiment,
            initial_soc=0.5,
            nproc=2,
        )



"""
Draw circuits with different terminal locations.
"""

import liionpack as lp

left = lp.setup_circuit(Np=3, Ns=1, terminals="left")
lp.draw_circuit(left)

right = lp.setup_circuit(Np=3, Ns=1, terminals="right")
lp.draw_circuit(right)

left_right = lp.setup_circuit(Np=3, Ns=1, terminals="left-right")
lp.draw_circuit(left_right)

right_left = lp.setup_circuit(Np=3, Ns=1, terminals="right-left")
lp.draw_circuit(right_left)

middle = lp.setup_circuit(Np=3, Ns=1, terminals=[1, 1])
lp.draw_circuit(middle)

lp.show_plots()



import liionpack as lp
import numpy as np
import pandas as pd
import pathlib
import pybamm
import unittest


class utilsTest(unittest.TestCase):
    def setUp(self):
        currents = [4, 5.2, 8, 10]
        volts = np.array([[3.5, 3.6], [3.54, 3.58]])
        output = {"currents": currents, "volts": volts}

        self.currents = currents
        self.volts = volts
        self.output = output

    def test_interp_current(self):
        d = {"Time": [0, 10], "Cells Total Current": [2.0, 4.0]}
        df = pd.DataFrame(data=d)
        f = lp.interp_current(df)
        assert f(5) == 3.0

    def test_add_events_to_model(self):
        model = pybamm.lithium_ion.SPMe()
        model = lp.add_events_to_model(model)
        events_in = False
        for key in sorted(model.variables.keys()):
            if "Event:" in key:
                events_in = True
                break
        assert events_in

    def test_build_inputs_dict(self):
        I_batt = np.array([1.0, 2.0])
        inputs = {"Electrode height [m]": [3.0, 4.0]}
        external_variables = {"Volume averaged cell temperature": [5.0, 6.0]}
        in_dict = lp.build_inputs_dict(I_batt, inputs, external_variables)
        assert len(in_dict) == 2

    def test_save_to_csv(self):
        lp.save_to_csv(self.output, path=".")

        with open("currents.csv", "r") as f:
            current = float(f.readline())
        self.assertEqual(self.currents[0], current)

        with open("volts.csv", "r") as f:
            volts = list(map(float, f.readline().split(", ")))
        self.assertEqual(self.volts[0][0], volts[0])

    def test_save_to_npy(self):
        lp.save_to_npy(self.output, path=".")

        currents = np.load("currents.npy")
        self.assertEqual(self.currents[0], currents[0])

        volts = np.load("volts.npy")
        self.assertEqual(self.volts[0, 0], volts[0, 0])

    def test_save_to_npz(self):
        lp.save_to_npzcomp(self.output, path=".")

        output = np.load("output.npz")
        currents = output["currents"]
        self.assertEqual(self.currents[0], currents[0])

        output = np.load("output.npz")
        volts = output["volts"]
        self.assertEqual(self.volts[0, 0], volts[0, 0])

    def tearDown(self):
        path = pathlib.Path("currents.csv")
        path.unlink(missing_ok=True)

        path = pathlib.Path("volts.csv")
        path.unlink(missing_ok=True)

        path = pathlib.Path("currents.npy")
        path.unlink(missing_ok=True)

        path = pathlib.Path("volts.npy")
        path.unlink(missing_ok=True)

        path = pathlib.Path("output.npz")
        path.unlink(missing_ok=True)


if __name__ == "__main__":
    unittest.main()



#
# Test solvers give the same output
#

import pybamm
import numpy as np
import unittest
import liionpack as lp


class TestSolvers(unittest.TestCase):
    def test_consistent_results_1_step(self):
        Rsmall = 1e-6
        netlist = lp.setup_circuit(
            Np=1, Ns=1, Rb=Rsmall, Rc=Rsmall, Ri=5e-2, V=4.0, I=1.0
        )
        # PyBaMM parameters
        parameter_values = pybamm.ParameterValues("Chen2020")
        # Cycling experiment
        experiment = pybamm.Experiment(
            [("Discharge at 1 A for 100 s or until 3.3 V",)] * 1, period="10 s"
        )
        # Solve pack with casadi
        a = lp.solve(
            netlist=netlist,
            parameter_values=parameter_values,
            experiment=experiment,
            inputs=None,
            nproc=1,
            manager="casadi",
        )
        # Solve pack with ray
        b = lp.solve(
            netlist=netlist,
            parameter_values=parameter_values,
            experiment=experiment,
            inputs=None,
            nproc=1,
            manager="ray",
        )

        v_a = a["Terminal voltage [V]"]
        v_b = b["Terminal voltage [V]"]

        assert np.allclose(v_a, v_b)


if __name__ == "__main__":
    unittest.main()



"""
Compare a drive-cycle simulation between PyBaMM and Liionpack.
"""

import pybamm
import os
import pandas as pd
import liionpack as lp
import matplotlib.pyplot as plt
import numpy as np

if __name__ == "__main__":
    lp.set_logging_level("NOTICE")

    os.chdir(pybamm.__path__[0] + "/..")
    netlist = lp.setup_circuit(Np=1, Ns=1, Rb=1.0e-6, Rc=1e-6, Ri=3e-2)

    parameter_values = pybamm.ParameterValues("Chen2020")

    # import drive cycle from file
    drive_cycle = pd.read_csv(
        pybamm.DataLoader().get_data("US06.csv"), comment="#", header=None
    ).to_numpy()

    timestep = 1
    drive_cycle[:, 0] *= timestep

    experiment = pybamm.Experiment(
        [pybamm.step.current(drive_cycle)],
        period=f"{timestep} seconds",
    )

    output_variables = [
        "X-averaged negative particle surface concentration [mol.m-3]",
        "X-averaged positive particle surface concentration [mol.m-3]",
    ]

    init_SoC = 0.5

    # Solve pack
    output = lp.solve(
        netlist=netlist,
        parameter_values=parameter_values,
        experiment=experiment,
        output_variables=output_variables,
        initial_soc=init_SoC,
        manager="casadi",
        nproc=8,
    )

    sim = pybamm.Simulation(
        model=pybamm.lithium_ion.SPM(),
        experiment=experiment,
        parameter_values=parameter_values,
    )

    sol = sim.solve(initial_soc=init_SoC)

    t_pybamm = sol["Time [s]"].entries
    t_liionpack = output["Time [s]"]
    v_pybamm = sol["Terminal voltage [V]"].entries
    v_liionpack = output["Terminal voltage [V]"]
    i_pybamm = sol["Current [A]"].entries
    i_liionpack = output["Cell current [A]"]
    r_pybamm = np.abs(sol["Local ECM resistance [Ohm]"].entries)
    r_liionpack = output["Cell internal resistance [Ohm]"]
    nconc_pybamm = sol[
        "X-averaged negative particle surface concentration [mol.m-3]"
    ].entries
    nconc_liionpack = output[
        "X-averaged negative particle surface concentration [mol.m-3]"
    ]
    pconc_pybamm = sol[
        "X-averaged positive particle surface concentration [mol.m-3]"
    ].entries
    pconc_liionpack = output[
        "X-averaged positive particle surface concentration [mol.m-3]"
    ]

    # The internal resistance in Liionpack is based on the previous time-step
    # so the solution lags behind. More advanced time stepping is required to
    # address this problem.
    t_liionpack -= timestep
    sol_diff = ((v_liionpack.flatten()[1:] - v_pybamm[:-1]) / v_pybamm[:-1]) * 100

    with plt.rc_context(lp.lp_context()):
        fig, ((ax0, ax1), (ax2, ax3), (ax4, ax5)) = plt.subplots(
            3, 2, figsize=(12, 10), sharex=True
        )
        ax0.plot(t_pybamm[:-1], sol_diff)
        ax0.set_ylabel("Voltage difference [%]")
        ax1.plot(t_pybamm, v_pybamm, label="PyBaMM")
        ax1.plot(t_liionpack, v_liionpack, "--", label="Liionpack")
        ax1.set_ylabel("Terminal voltage [V]")
        ax2.plot(t_pybamm, i_pybamm, label="PyBaMM")
        ax2.plot(t_liionpack, i_liionpack, "--", label="Liionpack")
        ax2.set_ylabel("Current [A]")
        ax3.plot(t_pybamm, r_pybamm, label="PyBaMM")
        ax3.plot(t_liionpack, r_liionpack, "--", label="Liionpack")
        ax3.set_ylabel("Internal resistance [Ohm]")
        ax4.plot(t_pybamm, nconc_pybamm, label="PyBaMM")
        ax4.plot(t_liionpack, nconc_liionpack, "--", label="Liionpack")
        ax4.set_xlabel("Time [s]")
        ax4.set_ylabel("Neg particle conc. [mol.m-3]")
        ax5.plot(t_pybamm, pconc_pybamm, label="PyBaMM")
        ax5.plot(t_liionpack, pconc_liionpack, "--", label="Liionpack")
        ax5.set_xlabel("Time [s]")
        ax5.set_ylabel("Pos particle conc. [mol.m-3]")
        handles, labels = ax5.get_legend_handles_labels()
        fig.legend(handles, labels, loc="upper right")

    plt.show()



import numpy as np
import pybamm


def generate_protocol_from_experiment(experiment):
    """

    Args:
        experiment (pybamm.Experiment):
            The experiment to generate the protocol from.

    Returns:
        protocol (list):
            a sequence of terminal currents to apply at each timestep
        terminations (list):
            a sequence voltage terminations for each step

    """
    protocol = []
    terminations = []
    step_types = []
    for i, step in enumerate(experiment.steps):
        proto = []
        t = step.duration
        dt = step.period
        termination = step.termination
        step_type = type(step).__name__.lower()
        if step_type not in ["current", "power"]:
            raise ValueError("Only current and power operations are supported")
        else:
            if not isinstance(step.value, pybamm.Interpolant):
                I = step.value
                proto.extend([I] * int(np.round(t, 5) / np.round(dt, 5)))
                if i == 0:
                    # Include initial state when not drive cycle, first op
                    proto = [proto[0]] + proto
            else:
                proto.extend(step.value.y.tolist())
            if len(termination) > 0:
                for term in termination:
                    if isinstance(
                        term, pybamm.experiment.step.step_termination.VoltageTermination
                    ):
                        terminations.append(term.value)
            else:
                terminations.append([])

        protocol.append(proto)
        step_types.append(step_type)

    return protocol, terminations, step_types



#
# Simulation utilities
#
import pybamm
import liionpack as lp
import numpy as np


def get_initial_stoichiometries(initial_soc, parameter_values):
    """
    Calculate initial stoichiometries to start off the simulation at a particular
    state of charge, given voltage limits, open-circuit potentials, etc defined by
    parameter_values

    Args:
        initial_soc (float):
            Target initial SOC. Must be between 0 and 1.
        parameter_values (pybamm.ParameterValues):
            The parameter values class that will be used for the simulation.
            Required for calculating appropriate initial stoichiometries.

    Returns:
        x, y (float):
            The initial stoichiometries that give the desired initial state of charge
    """
    if np.any(initial_soc < 0) or np.any(initial_soc > 1):
        raise ValueError("Initial SOC should be between 0 and 1")

    param = pybamm.LithiumIonParameters()
    esoh_solver = pybamm.lithium_ion.ElectrodeSOHSolver(parameter_values, param)
    return esoh_solver.get_initial_stoichiometries(initial_soc)


def update_init_conc(param, SoC=None, update=True):
    """
    Update initial concentration parameters

    Args:
        param (pybamm.ParameterValues):
            The battery simulation parameters.
        SoC (float):
            Target initial SoC. Must be between 0 and 1. Default is -1, in which
            case the initial concentrations are set using the target OCV.
        update (bool):
            Update the initial concentrations in place if True

    Returns:
        c_s_n_init (float):
            initial concentrations in negative particles.
        c_s_p_init (float):
            initial concentrations in positive particles.
    """
    c_n_max = param["Maximum concentration in negative electrode [mol.m-3]"]
    c_p_max = param["Maximum concentration in positive electrode [mol.m-3]"]
    x, y = lp.get_initial_stoichiometries(SoC, param)
    if x is not None:
        c_s_n_init, c_s_p_init = x * c_n_max, y * c_p_max
    else:
        return x, y
    if update:
        param.update(
            {
                "Initial concentration in negative electrode [mol.m-3]": c_s_n_init,
                "Initial concentration in positive electrode [mol.m-3]": c_s_p_init,
            }
        )
    return c_s_n_init, c_s_p_init



import liionpack as lp
import pybamm
import numpy as np
import matplotlib.pyplot as plt
import unittest


class solver_utilsTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        Np = 16
        Ns = 2
        self.Nspm = Np * Ns
        R_bus = 1e-4
        R_series = 1e-2
        R_int = 5e-2
        I_app = 80.0
        ref_voltage = 3.2
        # Generate the netlist
        self.netlist = lp.setup_circuit(
            Np, Ns, Rb=R_bus, Rc=R_series, Ri=R_int, V=ref_voltage, I=I_app
        )

        # Cycling experiment
        self.experiment = pybamm.Experiment(
            [
                f"Charge at {I_app} A for 300 seconds",
            ],
            period="10 seconds",
        )
        # PyBaMM parameters
        self.parameter_values = pybamm.ParameterValues("Chen2020")

    def test_mapped_step(self):
        pass

    def test_create_casadi_objects(self):
        pass

    def test_solve(self):
        output1 = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values,
            experiment=self.experiment,
            output_variables=None,
            inputs=None,
            initial_soc=0.5,
            nproc=1,
        )
        output2 = lp.solve(
            netlist=self.netlist.copy(),
            parameter_values=self.parameter_values,
            experiment=self.experiment,
            output_variables=None,
            inputs=None,
            initial_soc=0.5,
            nproc=2,
        )
        a = output1["Terminal voltage [V]"]
        b = output2["Terminal voltage [V]"]
        self.assertEqual(a.shape, (31, 32))
        self.assertTrue(np.allclose(a, b))

        plt.close("all")

    def test_solve_output_variables(self):
        var = "X-averaged negative particle surface concentration [mol.m-3]"
        output_variables = [
            var,
        ]
        output = lp.solve(
            netlist=self.netlist,
            parameter_values=self.parameter_values,
            experiment=self.experiment,
            output_variables=output_variables,
            inputs=None,
            initial_soc=0.5,
        )
        self.assertEqual(output[var].shape, (31, 32))
        plt.close("all")

    def test_sim_func(self):
        def bespoke_sim(parameter_values):
            model = pybamm.lithium_ion.SPM(
                options={
                    "thermal": "lumped",
                }
            )
            parameter_values.update(
                {
                    "Current function [A]": "[input]",
                    "Total heat transfer coefficient [W.m-2.K-1]": "[input]",
                },
            )

            # Set up solver and simulation
            solver = pybamm.CasadiSolver(mode="safe")
            sim = pybamm.Simulation(
                model=model,
                experiment=None,
                parameter_values=parameter_values,
                solver=solver,
            )
            return sim

        # Heat transfer coefficients
        inputs = {
            "Total heat transfer coefficient [W.m-2.K-1]": np.ones(self.Nspm) * 10
        }
        output = lp.solve(
            netlist=self.netlist,
            sim_func=bespoke_sim,
            parameter_values=self.parameter_values,
            experiment=self.experiment,
            output_variables=None,
            inputs=inputs,
            initial_soc=0.5,
        )
        self.assertEqual(output["Terminal voltage [V]"].shape, (31, 32))
        plt.close("all")


if __name__ == "__main__":
    unittest.main()



import liionpack as lp
import matplotlib.pyplot as plt
import pybamm
import unittest
import numpy as np


class plotsTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        R_bus = 1e-4
        R_series = 1e-2
        R_int = 5e-2
        I_app = 1.0
        ref_voltage = 3.2

        # Load the netlist
        self.netlist = lp.read_netlist(
            "4p1s", Ri=R_int, Rc=R_series, Rb=R_bus, Rt=R_bus, I=I_app, V=ref_voltage
        )

        # Cycling experiment
        experiment = pybamm.Experiment(
            [
                f"Charge at {I_app} A for 300 seconds",
                "Rest for 100 seconds",
                f"Discharge at {I_app} A for 300 seconds",
                "Rest for 100 seconds",
            ],
            period="10 seconds",
        )
        # PyBaMM parameters
        parameter_values = pybamm.ParameterValues("Chen2020")
        # Solve pack
        output = lp.solve(
            netlist=self.netlist,
            parameter_values=parameter_values.copy(),
            experiment=experiment,
            output_variables=None,
            inputs=None,
            initial_soc=0.5,
        )
        self.output = output
        self.sim = pybamm.Simulation(
            model=pybamm.lithium_ion.SPM(),
            parameter_values=parameter_values.copy(),
            experiment=experiment,
        )

    def test_draw_circuit(self):
        net = lp.setup_circuit(
            Np=3, Ns=1, Rb=1e-4, Rc=1e-2, Ri=5e-2, V=3.2, I=80.0, terminals=[0, 1]
        )
        lp.draw_circuit(net)
        plt.close("all")

    def test_plot_pack(self):
        lp.plot_pack(self.output)
        plt.close("all")

    def test_plot_cells(self):
        lp.plot_cells(self.output)
        plt.close("all")

    def test_plot_output(self):
        lp.plot_output(self.output)
        plt.close("all")
        lp.plot_output(self.output, color="white")
        plt.close("all")

    def test_compare_plots(self):
        solution = self.sim.solve(initial_soc=0.5)
        lp.compare_solution_output(solution, self.output)
        lp.compare_solution_output(self.output, solution)
        plt.close("all")

    def test_plot_cell_data_image(self):
        Np = 5
        Ns = 2
        Nspm = Np * Ns
        # Generate the netlist
        netlist = lp.setup_circuit(Np=Np, Ns=Ns, Rb=1e-4, Ri=3e-2)
        # Define additional output variables
        output_variables = [
            "Volume-averaged cell temperature [K]",
        ]
        # Define a cycling experiment using PyBaMM
        experiment = pybamm.Experiment(
            [
                "Discharge at 5 A for 5 minutes",
            ],
            period="10 seconds",
        )
        # Define the PyBaMM parameters
        parameter_values = pybamm.ParameterValues("Chen2020")
        parameter_values.update(
            {"Total heat transfer coefficient [W.m-2.K-1]": "[input]"}
        )
        htc = np.random.random(Nspm) * 10.0
        inputs = {"Total heat transfer coefficient [W.m-2.K-1]": htc}
        # Solve the pack
        output = lp.solve(
            netlist=netlist,
            sim_func=lp.thermal_simulation,
            parameter_values=parameter_values,
            experiment=experiment,
            output_variables=output_variables,
            inputs=inputs,
            initial_soc=0.5,
        )
        data = output["Volume-averaged cell temperature [K]"][-1, :]
        lp.plot_cell_data_image(netlist, data)
        plt.close("all")


if __name__ == "__main__":
    unittest.main()



"""
A basic example of a pack simulation consisting of two sets of 16 parallel
cells connected in series for a total of 32 cells.
"""

import liionpack as lp
import pybamm
import numpy as np
import os

lp.set_logging_level("NOTICE")

# Define parameters
Np = 16
Ns = 2
Iapp = 20

# Generate the netlist
netlist = lp.setup_circuit(Np=Np, Ns=Ns)

# Define additional output variables
output_variables = ["Volume-averaged cell temperature [K]"]

# Define a cycling experiment using PyBaMM
experiment = pybamm.Experiment(
    [
        f"Charge at {Iapp} A for 30 minutes",
        "Rest for 15 minutes",
        f"Discharge at {Iapp} A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# Define the PyBaMM parameters
parameter_values = pybamm.ParameterValues("Chen2020")
inputs = {"Total heat transfer coefficient [W.m-2.K-1]": np.ones(Np * Ns) * 10}

# Solve the pack
output = lp.solve(
    netlist=netlist,
    sim_func=lp.thermal_simulation,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=0.5,
    inputs=inputs,
    nproc=os.cpu_count(),
    manager="casadi",
)

# Plot the pack and individual cell results
lp.plot_pack(output)
lp.plot_cells(output)
lp.show_plots()



#
# Test some experiments
#
import pybamm
import numpy as np
import unittest
import liionpack as lp


class Test1p1s(unittest.TestCase):
    def test_consistent_results_1_step(self):
        Rsmall = 1e-6
        netlist = lp.setup_circuit(
            Np=1, Ns=1, Rb=Rsmall, Rc=Rsmall, Ri=5e-2, V=4.0, I=1.0
        )
        # PyBaMM parameters
        parameter_values = pybamm.ParameterValues("Chen2020")
        # Cycling experiment
        experiment = pybamm.Experiment(
            [("Discharge at 1 A for 100 s or until 3.3 V",)] * 1, period="10 s"
        )
        # Solve pack
        output = lp.solve(
            netlist=netlist, parameter_values=parameter_values, experiment=experiment
        )

        parameter_values = pybamm.ParameterValues("Chen2020")
        sim = pybamm.Simulation(
            model=pybamm.lithium_ion.SPM(),
            parameter_values=parameter_values,
            experiment=experiment,
        )

        sol = sim.solve()
        a = output["Terminal voltage [V]"].flatten()
        b = sol["Terminal voltage [V]"].entries

        assert np.allclose(a, b)

    def test_consistent_results_2_step(self):
        Rsmall = 1e-6
        netlist = lp.setup_circuit(
            Np=1, Ns=1, Rb=Rsmall, Rc=Rsmall, Ri=5e-2, V=4.0, I=1.0
        )
        # PyBaMM parameters
        parameter_values = pybamm.ParameterValues("Chen2020")
        # Cycling experiment
        experiment = pybamm.Experiment(
            [
                (
                    "Discharge at 1 A for 100 s or until 3.3 V",
                    "Rest for 100 s",
                )
            ]
            * 1,
            period="10 s",
        )
        SoC = 0.5
        # Solve pack
        output = lp.solve(
            netlist=netlist,
            parameter_values=parameter_values,
            experiment=experiment,
            initial_soc=SoC,
        )

        parameter_values = pybamm.ParameterValues("Chen2020")
        sim = pybamm.Simulation(
            model=pybamm.lithium_ion.SPM(),
            parameter_values=parameter_values,
            experiment=experiment,
        )

        sol = sim.solve(initial_soc=SoC)
        a = output["Terminal voltage [V]"].flatten()
        b = sol["Terminal voltage [V]"].entries
        diff = np.abs(a[:20] - b[:20])
        assert np.all(diff < 0.05)


if __name__ == "__main__":
    unittest.main()



#
# External thermal example
#

import liionpack as lp
import pybamm
import numpy as np
import matplotlib.pyplot as plt

plt.close("all")


# Generate the netlist
netlist = lp.setup_circuit(Np=4, Ns=1, Rb=1e-3, Rc=1e-2)

# Define some additional variables to output
output_variables = [
    "Volume-averaged cell temperature [K]",
    "Volume-averaged total heating [W.m-3]",
]

# Cycling experiment, using PyBaMM
experiment = pybamm.Experiment(["Discharge at 5 A for 5 minutes"], period="10 seconds")

# PyBaMM battery parameters
parameter_values = pybamm.ParameterValues("Chen2020")

# Solve the pack problem
temps = np.ones(4) * 300 + np.arange(4) * 10
inputs = {"Input temperature [K]": temps}
output = lp.solve(
    netlist=netlist,
    sim_func=lp.thermal_external,
    inputs=inputs,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=0.5,
)

# Display the results
lp.plot_output(output, color="white")



import liionpack as lp
import pybamm
import unittest


class sim_utilsTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.param = pybamm.ParameterValues("Chen2020")

    def test_update_init_conc(self):
        a = self.param["Initial concentration in negative electrode [mol.m-3]"]
        lp.update_init_conc(self.param, SoC=0.0, update=False)
        b = self.param["Initial concentration in negative electrode [mol.m-3]"]
        lp.update_init_conc(self.param, SoC=0.0, update=True)
        c = self.param["Initial concentration in negative electrode [mol.m-3]"]
        assert a == b
        assert a > c

    def test_bad_soc(self):
        with self.assertRaises(ValueError):
            lp.update_init_conc(self.param, SoC=10.0)

    def test_input_logger_warning(self):
        with self.assertRaises(ValueError):
            param = pybamm.ParameterValues("Chen2020")
            neg_conc = 0.0
            param.update(
                {
                    "Electrode height [m]": "[input]",
                    "Initial concentration in negative electrode [mol.m-3]": neg_conc,
                },
                check_already_exists=False,
            )
            lp.update_init_conc(param, SoC=0.5, update=True)
            # a = param["Initial concentration in negative electrode [mol.m-3]"]
            # assert a == neg_conc


if __name__ == "__main__":
    unittest.main()



import numpy as np
import codecs
import pandas as pd
import liionpack as lp
import os
import pybamm
import scipy as sp
from lcapy import Circuit


def read_netlist(
    filepath,
    Ri=None,
    Rc=None,
    Rb=None,
    Rt=None,
    I=None,
    V=None,
):
    """
    Assumes netlist has been saved by LTSpice with format Descriptor Node1 Node2 Value
    Any lines starting with * are comments and . are commands so ignore them
    Nodes begin with N so remove that
    Open ended components are not allowed and their nodes start with NC (no-connection)

    Args:
        filepath (str): Path to netlist circuit file '.cir' or '.txt'.
        Ri (float): Internal resistance ($\Omega$).
        Rc (float): Connection resistance ($\Omega$).
        Rb (float): Busbar resistance ($\Omega$).
        Rt (float): Terminal connection resistance ($\Omega$).
        I (float): Current (A).
        V (float): Initial battery voltage (V).

    Returns:
        netlist (pandas.DataFrame):
            A netlist of circuit elements with format desc, node1, node2, value.
    """

    # Read in the netlist
    if "." not in filepath:
        filepath += ".cir"
    if not os.path.isfile(filepath):
        temp = os.path.join(lp.CIRCUIT_DIR, filepath)
        if not os.path.isfile(temp):
            pass
        else:
            filepath = temp
    if ".cir" in filepath:
        with codecs.open(filepath, "r", "utf-16LE") as fd:
            Lines = fd.readlines()
    elif ".txt" in filepath:
        with open(filepath, "r") as f:
            Lines = f.readlines()
    else:
        raise FileNotFoundError(
            'Please supply a valid file with extension ".cir" or ".txt"'
        )
    # Ignore lines starting with * or .
    Lines = [l.strip("\n").split(" ") for l in Lines if l[0] not in ["*", "."]]
    Lines = np.array(Lines, dtype="<U16")

    # Read descriptions and nodes, strip N from nodes
    # Lines is desc | node1 | node2
    desc = Lines[:, 0]
    node1 = Lines[:, 1]
    node2 = Lines[:, 2]
    value = Lines[:, 3]
    try:
        value = value.astype(float)
    except ValueError:
        pass
    node1 = np.array([x.strip("N") for x in node1], dtype=int)
    node2 = np.array([x.strip("N") for x in node2], dtype=int)
    netlist = pd.DataFrame(
        {"desc": desc, "node1": node1, "node2": node2, "value": value}
    )

    # Populate the values based on the descriptions (element types)
    for name, val in [
        ("Ri", Ri),
        ("Rc", Rc),
        ("Rb", Rb),
        ("Rl", Rb),
        ("Rt", Rt),
        ("I", I),
        ("V", V),
    ]:
        if val is not None:
            # netlist["desc"] consists of entries like 'Ri13'
            # this map finds all the entries that start with (e.g.) 'Ri'
            name_map = netlist["desc"].str.find(name) > -1
            # then allocates the value to the corresponding indices
            netlist.loc[name_map, ("value")] = val

    lp.logger.notice("netlist " + filepath + " loaded")
    return netlist


def setup_circuit(
    Np=1,
    Ns=1,
    Ri=1e-2,
    Rc=1e-2,
    Rb=1e-4,
    Rt=1e-5,
    I=80.0,
    V=4.2,
    plot=False,
    terminals="left",
    configuration="parallel-strings",
):
    """
    Define a netlist from a number of batteries in parallel and series

    Args:
        Np (int): Number of batteries in parallel.
        Ns (int): Number of batteries in series.
        Ri (float): Internal resistance ($\Omega$).
        Rc (float): Connection resistance ($\Omega$).
        Rb (float): Busbar resistance ($\Omega$).
        Rt (float): Terminal connection resistance ($\Omega$).
        I (float): Current (A).
        V (float): Initial battery voltage (V).
        plot (bool): Plot the circuit.
        terminals (string): The location of the terminals. Can be "left", "right",
            "left-right", "right-left" or a list or array of node integers.
        configuration (string): The pack circuit configuration to use. Can be
            "parallel-strings" (default) or "series-groups"

    Returns:
        netlist (pandas.DataFrame):
            A netlist of circuit elements with format desc, node1, node2, value.

    """
    Nc = Np
    Nr = Ns * 3 + 1

    grid = np.arange(Nc * Nr).reshape([Nr, Nc])
    coords = np.indices(grid.shape)
    y = coords[0, :, :]
    x = coords[1, :, :]
    # make contiguous now instead of later when netlist is done as very slow
    mask = np.ones([Nr, Nc], dtype=bool)
    # This is no longer needed as terminals connect directly to battery
    # Guess could also add a terminal connection resistor though
    # mask[1:-1, 0] = False
    grid[mask] = np.arange(np.sum(mask)) + 1
    x = x[mask].flatten()
    y = y[mask].flatten()
    grid[~mask] = -2  # These should never be used

    # grid is a Nr x Nc matrix
    # 1st column is terminals only
    # 1st and last rows are busbars
    # Other rows alternate between series resistor and voltage source
    # For example if Np=1 and Nc=2,
    # grid = array([[ 0,  1], # busbar
    #                         # Rs
    #               [ 2,  3],
    #                         # V
    #               [ 4,  5],
    #                         # Ri
    #               [ 6,  7],
    #                         # Rs
    #               [ 8,  9],
    #                         # V
    #               [10, 11],
    #                         # Ri
    #               [12, 13]] # busbar)
    # Connections are across busbars in first and last rows, and down each column
    # See "01 Getting Started.ipynb"

    # Build data  with ['element type', node1, node2, value]
    netlist = []

    num_Rb = 0
    num_V = 0

    desc = []
    node1 = []
    node2 = []
    value = []

    # -ve busbars (bottom row of the grid)
    bus_nodes = [grid[0, :]]
    for nodes in bus_nodes:
        for i in range(len(nodes) - 1):
            # netline = []
            desc.append("Rbn" + str(num_Rb))
            num_Rb += 1
            node1.append(nodes[i])
            node2.append(nodes[i + 1])
            value.append(Rb)
    num_Rs = 0
    num_Ri = 0
    # Series resistors and voltage sources
    cols = np.arange(Nc)
    rows = np.arange(Nr)[:-1]
    rtype = ["Rc", "V", "Ri"] * Ns
    for col in cols:
        # Go down the column alternating Rs, V, Ri connections between nodes
        nodes = grid[:, col]
        for row in rows:
            if rtype[row] == "Rc":
                # Inter(c)onnection / weld
                desc.append(rtype[row] + str(num_Rs))
                num_Rs += 1
                val = Rc
            elif rtype[row] == "Ri":
                # Internal resistor
                desc.append(rtype[row] + str(num_Ri))
                num_Ri += 1
                val = Ri
            else:
                # Voltage source
                desc.append("V" + str(num_V))
                num_V += 1
                val = V
            node1.append(nodes[row + 1])
            node2.append(nodes[row])
            value.append(val)
            # netlist.append(netline)

    # +ve busbar (top row of the grid)
    if configuration == "parallel-strings":
        bus_nodes = [grid[-1, :]]
    elif configuration == "series-groups":
        bus_nodes = grid[3::3, :]
    else:
        raise ValueError("configuration must be parallel-strings or series-groups")

    for nodes in bus_nodes:
        for i in range(len(nodes) - 1):
            # netline = []
            desc.append("Rbp" + str(num_Rb))
            num_Rb += 1
            node1.append(nodes[i])
            node2.append(nodes[i + 1])
            value.append(Rb)

    desc = np.asarray(desc)
    node1 = np.asarray(node1)
    node2 = np.asarray(node2)
    value = np.asarray(value)
    main_grid = {
        "desc": desc,
        "node1": node1,
        "node2": node2,
        "value": value,
        "node1_x": x[node1 - 1],
        "node1_y": y[node1 - 1],
        "node2_x": x[node2 - 1],
        "node2_y": y[node2 - 1],
    }

    # Current source - spans the entire pack
    if (terminals == "left") or (terminals is None):
        t_nodes = [0, 0]
    elif terminals == "right":
        t_nodes = [-1, -1]
    elif terminals == "left-right":
        t_nodes = [0, -1]
    elif terminals == "right-left":
        t_nodes = [-1, 0]
    elif isinstance(terminals, (list, np.ndarray)):
        t_nodes = terminals
    else:
        raise ValueError(
            'Please specify a valid terminals argument: "left", '
            + '"right", "left-right" or "right-left" or a list or '
            + "array of nodes"
        )
    # terminal nodes
    t1 = grid[-1, t_nodes[0]]
    t2 = grid[0, t_nodes[1]]
    # terminal coords
    x1 = x[t1 - 1]
    x2 = x[t2 - 1]
    y1 = y[t1 - 1]
    y2 = y[t2 - 1]
    nn = grid.max() + 1  # next node
    # coords of nodes forming current source loop
    if terminals == "left" or (
        isinstance(terminals, (list, np.ndarray)) and np.all(np.array(terminals) == 0)
    ):
        ix = x1 - 1
        dy = 0
    elif terminals == "right" or (
        isinstance(terminals, (list, np.ndarray)) and np.all(np.array(terminals) == -1)
    ):
        ix = x1 + 1
        dy = 0
    else:
        ix = -1
        dy = 1
    if dy == 0:
        desc = ["Rtp1", "I0", "Rtn1"]
        xs = np.array([x1, ix, ix, x2])
        ys = np.array([y1, y1, y2, y2])
        node1 = [t1, nn, 0]
        node2 = [nn, 0, t2]
        value = [Rt, I, Rt]
        num_elem = 3
    else:
        desc = ["Rtp0", "Rtp1", "I0", "Rtn1", "Rtn0"]
        xs = np.array([x1, x1, ix, ix, x2, x2])
        ys = np.array([y1, y1 + dy, y1 + dy, 0 - dy, 0 - dy, y2])
        node1 = [t1, nn, nn + 1, 0, nn + 2]
        node2 = [nn, nn + 1, 0, nn + 2, t2]
        hRt = Rt / 2
        value = [hRt, hRt, I, hRt, hRt]
        num_elem = 5

    desc = np.asarray(desc)
    node1 = np.asarray(node1)
    node2 = np.asarray(node2)
    value = np.asarray(value)
    current_loop = {
        "desc": desc,
        "node1": node1,
        "node2": node2,
        "value": value,
        "node1_x": xs[:num_elem],
        "node1_y": ys[:num_elem],
        "node2_x": xs[1:],
        "node2_y": ys[1:],
    }

    for key in main_grid.keys():
        main_grid[key] = np.concatenate((main_grid[key], current_loop[key]))
    netlist = pd.DataFrame(main_grid)

    if plot:
        lp.simple_netlist_plot(netlist)
    lp.logger.notice("Circuit created")
    return netlist


def solve_circuit(netlist, current=None, power=None):
    """
    Generate and solve the Modified Nodal Analysis (MNA) equations for the circuit.
    The MNA equations are a linear system Ax = z.
    See http://lpsa.swarthmore.edu/Systems/Electrical/mna/MNA3.html

    Args:
        netlist (pandas.DataFrame):
            A netlist of circuit elements with format desc, node1, node2, value.
        current (float):
            The current value for the current source. Overrides the netlist value.
        power (float):
            The power value for the power source. Overrides the netlist value.

    Returns:
        V_node (np.ndarray):
            Voltages of the voltage elements.
        I_batt (np.ndarray):
            Currents of the current elements.
        terminal_voltage (float):
            Voltage at the terminal.
        terminal_current (float):
            Current at the terminal.
        terminal_power (float):
            Power at the terminal.
    """

    control_args = sum(arg is not None for arg in [current, power])
    if control_args == 2:
        raise ValueError("Only specify one of current or power arguments.")

    timer = pybamm.Timer()

    desc = np.array(netlist["desc"]).astype("<U1")  # just take first character
    I_map = desc == "I"
    R_map = desc == "R"
    V_map = desc == "V"
    node1 = np.array(netlist["node1"])
    node2 = np.array(netlist["node2"])
    value = np.array(netlist["value"])
    n = np.concatenate((node1, node2)).max()  # Number of nodes (highest node number)

    m = np.sum(desc == "V")  # we only use V in liionpack

    # Construct the A matrix, which will be a (n+m) x (n+m) matrix
    G = sp.sparse.lil_matrix((n, n))
    B = sp.sparse.lil_matrix((n, m))
    D = sp.sparse.lil_matrix((m, m))
    i = np.zeros([n, 1])
    e = np.zeros([m, 1])

    node1 = node1 - 1  # get the two node numbers in python index format
    node2 = node2 - 1
    # Resistance elements: fill the G matrix only
    g = np.ones(len(value)) * np.nan
    n1_ground = node1 == -1
    n2_ground = node2 == -1
    # Resistors

    g[R_map] = 1 / value[R_map]  # conductance = 1 / R
    R_map_n1_ground = np.logical_and(R_map, n1_ground)
    R_map_n2_ground = np.logical_and(R_map, n2_ground)
    R_map_ok = np.logical_and(R_map, ~np.logical_or(n1_ground, n2_ground))

    if np.any(R_map_n1_ground):  # -1 is the ground node
        n2 = node2[R_map_n1_ground]
        G[n2, n2] = G[n2, n2] + g[R_map_n1_ground]
    if np.any(R_map_n2_ground):
        n1 = node1[R_map_n2_ground]
        G[n1, n1] = G[n1, n1] + g[R_map_n2_ground]

    n1 = node1[R_map_ok]
    n2 = node2[R_map_ok]
    g_change = g[R_map_ok]
    gn1n1 = sp.sparse.csr_matrix((g_change, (n1, n1)), shape=G.shape)
    gn2n2 = sp.sparse.csr_matrix((g_change, (n2, n2)), shape=G.shape)
    gn1n2 = sp.sparse.csr_matrix((g_change, (n1, n2)), shape=G.shape)
    gn2n1 = sp.sparse.csr_matrix((g_change, (n2, n1)), shape=G.shape)
    G += gn1n1
    G += gn2n2
    G -= gn1n2
    G -= gn2n1

    V_map_not_n1_ground = np.logical_and(V_map, ~n1_ground)
    V_map_not_n2_ground = np.logical_and(V_map, ~n2_ground)
    n1 = node1[V_map_not_n1_ground]
    n2 = node2[V_map_not_n2_ground]
    vsCnt = np.ones(len(V_map)) * -1
    vsCnt[V_map] = np.arange(m)
    B[n1, vsCnt[V_map_not_n1_ground]] = 1
    B[n2, vsCnt[V_map_not_n2_ground]] = -1
    e[np.arange(m), 0] = value[V_map]

    upper = sp.sparse.hstack((G, B))
    lower = sp.sparse.hstack((B.T, D))
    A = sp.sparse.vstack((upper, lower))
    A_csr = sp.sparse.csr_matrix(A)

    n1 = node1[I_map]
    n2 = node2[I_map]

    toc_setup = timer.time()
    lp.logger.debug(f"Circuit set up in {toc_setup}")

    def _solve(A_csr, z, n):
        X = sp.sparse.linalg.spsolve(A_csr, z).flatten()
        V_node = np.zeros(n + 1)
        V_node[1:] = X[:n]
        I_batt = X[n:]
        return V_node, I_batt

    Terminal_Node = np.array(netlist[I_map].node1)

    # If no control arguments are specified, solve the circuit with the current values
    if control_args == 0:
        current = value[I_map]
    if current is not None:
        if n1 >= 0:
            i[n1] = i[n1] - current
        if n2 >= 0:
            i[n2] = i[n2] + current
        z = np.vstack((i, e))
        V_node, I_batt = _solve(A_csr, z, n)
        netlist.loc[I_map, ("value")] = current

    elif power is not None:
        current_guess = value[I_map]
        max_iterations = 100
        iteration = 0
        tolerance = 0.001
        while iteration < max_iterations:
            i = np.zeros([n, 1])
            if n1 >= 0:
                i[n1] = i[n1] - current_guess
            if n2 >= 0:
                i[n2] = i[n2] + current_guess
            z = np.vstack((i, e))
            V_node, I_batt = _solve(A_csr, z, n)
            V_Terminal = V_node[Terminal_Node]
            power_guess = V_Terminal * current_guess
            if abs(power_guess - power) < tolerance:
                break
            current_adjustment = (power - power_guess) / V_Terminal
            current_guess += current_adjustment
            iteration += 1
        netlist.loc[I_map, ("value")] = current_guess

    terminal_voltage = V_node[Terminal_Node]
    terminal_current = netlist.loc[I_map].value.to_numpy()
    terminal_power = terminal_voltage * terminal_current

    toc = timer.time()
    lp.logger.debug(f"Circuit solved in {toc - toc_setup}")
    lp.logger.info(f"Circuit set up and solved in {toc}")
    return V_node, I_batt, terminal_current, terminal_voltage, terminal_power


def make_lcapy_circuit(netlist):
    """
    Generate a circuit that can be used with lcapy

    Args:
        netlist (pandas.DataFrame):
            A netlist of circuit elements with format. desc, node1, node2, value.

    Returns:
        cct (lcapy.Circuit):
            The Circuit class is used for describing networks using netlists.
            Despite the name, it does not require a closed path.

    """
    cct = Circuit()
    I_map = netlist["desc"].str.find("I") > -1
    net2 = netlist.copy()
    net2.loc[I_map, ("node1")] = netlist["node2"][I_map]
    net2.loc[I_map, ("node2")] = netlist["node1"][I_map]
    d1 = "down"
    d2 = "up"
    I_xs = [net2[I_map]["node1_x"].values[0], net2[I_map]["node2_x"].values[0]]
    I_left = np.any(np.array(I_xs) == -1)
    all_desc = netlist["desc"].values
    for index, row in net2.iterrows():
        color = "black"
        desc, n1, n2, value, n1x, n1y, n2x, n2y = row[:8]
        if desc[0] == "V":
            direction = d1
        elif desc[0] == "I":
            direction = d2
        elif desc[0] == "R":
            if desc[1] == "b":
                direction = "right"
            elif desc[1] == "t":
                # These are the terminal nodes and require special attention
                if desc[2] == "p":
                    # positive
                    color = "red"
                else:
                    # negative
                    color = "blue"
                # If terminals are not both at the same end then the netlist
                # has two resistors with half the value to make a nice circuit
                # diagram. Convert into 1 resistor + 1 wire
                if desc[3] == "0":
                    # The wires have the zero suffix
                    direction = d2
                    desc = "W"
                else:
                    # The reistors have the 1 suffix
                    # Convert the value to the total reistance if a wire element
                    # is in the netlist
                    w_desc = desc[:3] + "0"
                    if w_desc in all_desc:
                        value *= 2
                    desc = desc[:3]
                    # Terminal loop is C shaped with positive at the top so
                    # order is left-vertical-right if we're on the left side
                    # and right-vertical-left if we're on the right side
                    if desc[2] == "p":
                        if I_left:
                            direction = "left"
                            # if the terminal connection is not at the end then
                            # extend the element connections
                            if n1x > 0:
                                direction += "=" + str(1 + n1x)
                        else:
                            direction = "right"
                            if n1x < I_xs[0] - 1:
                                direction += "=" + str(1 + I_xs[0] - n1x)
                    else:
                        if I_left:
                            direction = "right"
                        else:
                            direction = "left"
            else:
                direction = d1
        if desc == "W":
            string = desc + " " + str(n1) + " " + str(n2)
        else:
            string = desc + " " + str(n1) + " " + str(n2) + " " + str(value)
        string = string + "; " + direction
        string = string + ", color=" + color
        cct.add(string)
    # Add ground node
    cct.add("W 0 00; down, sground")
    return cct


def power_loss(netlist, include_Ri=False):
    """
    Calculate the power loss through joule heating of all the resistors in the
    circuit

    Args:
        netlist (pandas.DataFrame):
            A netlist of circuit elements with format desc, node1, node2, value.
        include_Ri (bool):
            Default is False. If True the internal resistance of the batteries
            is included

    """
    V_node, I_batt, t_c, t_v, t_p = lp.solve_circuit(netlist)
    R_map = netlist["desc"].str.find("R") > -1
    R_map = R_map.values
    if not include_Ri:
        Ri_map = netlist["desc"].str.find("Ri") > -1
        Ri_map = Ri_map.values
        R_map *= ~Ri_map
    R_value = netlist[R_map].value.values
    R_node1 = netlist[R_map].node1.values
    R_node2 = netlist[R_map].node2.values
    R_node1_V = V_node[R_node1]
    R_node2_V = V_node[R_node2]
    V_diff = np.abs(R_node1_V - R_node2_V)
    P_loss = V_diff**2 / R_value
    netlist["power_loss"] = 0.0
    netlist.loc[R_map, ("power_loss")] = P_loss


def _fn(n):
    if n == 0:
        return "0"
    else:
        return "N" + str(n).zfill(3)


def write_netlist(netlist, filename):
    """
    Write netlist to file

    Args:
        netlist (pandas.DataFrame):
            A netlist of circuit elements with format desc, node1, node2, value.

    """
    lines = ["* " + filename]
    for i, r in netlist.iterrows():
        line = r.desc + " " + _fn(r.node1) + " " + _fn(r.node2) + " " + str(r.value)
        lines.append(line)
    lines.append(".op")
    lines.append(".backanno")
    lines.append(".end")
    with open(filename, "w") as f:
        for line in lines:
            f.write(line)
            f.write("\n")



#
# Paper example
#

import liionpack as lp
import pybamm

# Generate the netlist
netlist = lp.setup_circuit(Np=4, Ns=1, Rb=1e-3, Rc=1e-2)

# Define some additional variables to output
output_variables = [
    "X-averaged negative particle surface concentration [mol.m-3]",
    "X-averaged positive particle surface concentration [mol.m-3]",
]

# Cycling experiment, using PyBaMM
experiment = pybamm.Experiment(
    [
        "Charge at 5 A for 30 minutes",
        "Rest for 15 minutes",
        "Discharge at 5 A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# PyBaMM battery parameters
parameter_values = pybamm.ParameterValues("Chen2020")

# Solve the pack problem
output = lp.solve(
    netlist=netlist,
    parameter_values=parameter_values,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=0.5,
)

# Display the results
lp.plot_output(output, color="white")

# Draw the circuit at final state
lp.draw_circuit(netlist, cpt_size=1.0, node_spacing=2.2)



"""
Example of running a simulation with batteries of different size.
"""

import liionpack as lp
import pybamm
import numpy as np

lp.logger.setLevel("NOTICE")

# Define parameters
Np = 2
Ns = 3
Iapp = 10

# Generate the netlist and output variables
netlist = lp.setup_circuit(
    Np=Np,
    Ns=Ns,
    Rb=1.5e-3,
    Rc=1e-2,
    Ri=5e-2,
    V=4.0,
    I=Iapp,
    configuration="series-groups",
)

lp.draw_circuit(netlist)

# Cycling experiment
experiment = pybamm.Experiment(
    [
        f"Discharge at {Iapp} A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# PyBaMM parameters
param = pybamm.ParameterValues("Chen2020")

w_original = param["Electrode width [m]"]

param.update(
    {
        "Electrode width [m]": "[input]",
    }
)

new_widths = np.ones(Np * Ns) * w_original

# Divide the capacity by 2 for 1 half of the batteries
new_widths[:3] = w_original / 2

inputs = {
    "Electrode width [m]": new_widths,
}

# Solve pack
output = lp.solve(
    netlist=netlist,
    parameter_values=param,
    experiment=experiment,
    initial_soc=None,
    inputs=inputs,
)

# Plot results
lp.plot_output(output)
lp.show_plots()



import liionpack as lp
import pybamm
import unittest


class simulationsTest(unittest.TestCase):
    def test_basic_simulation(self):
        sim = lp.basic_simulation()
        sim.solve([0, 1800])
        assert sim.__class__ == pybamm.Simulation

    def test_thermal_simulation(self):
        sim = lp.thermal_simulation()
        sim.solve(
            [0, 1800], inputs={"Total heat transfer coefficient [W.m-2.K-1]": 1.0}
        )
        assert sim.__class__ == pybamm.Simulation

    # def test_thermal_external(self):
    #     sim = lp.thermal_external()
    #     sim.solve([0, 1800],
    #               external_variables={"Volume-averaged cell temperature": 1.0})
    #     assert sim.__class__ == pybamm.Simulation


if __name__ == "__main__":
    unittest.main()



"""
Set up a simulation and draw the circuit.
"""

import liionpack as lp
import pybamm

lp.logger.setLevel("NOTICE")

# Define parameters
Np = 4
Ns = 1

# Generate the netlist and draw circuit
netlist = lp.setup_circuit(Np=Np, Ns=Ns, I=5)
lp.draw_circuit(netlist, node_spacing=2.5)

output_variables = [
    "X-averaged negative particle surface concentration [mol.m-3]",
    "X-averaged positive particle surface concentration [mol.m-3]",
]

# Cycling experiment
experiment = pybamm.Experiment(
    [
        "Charge at 5 A for 30 minutes",
        "Rest for 15 minutes",
        "Discharge at 5 A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# PyBaMM parameters
parameter_values = pybamm.ParameterValues("Chen2020")

# Solve pack
output = lp.solve(
    netlist=netlist,
    parameter_values=parameter_values,
    experiment=experiment,
    initial_soc=0.5,
    output_variables=output_variables,
)

lp.plot_output(output)
lp.show_plots()



import liionpack as lp
import pybamm
import numpy as np
import unittest


class solversTest(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        Np = 21
        Ns = 1
        self.Nspm = Np * Ns
        R_bus = 1e-4
        R_series = 1e-2
        R_int = 5e-2
        I_app = 20.0
        ref_voltage = 3.2
        # Generate the netlist
        self.netlist = lp.setup_circuit(
            Np, Ns, Rb=R_bus, Rc=R_series, Ri=R_int, V=ref_voltage, I=I_app
        )

        # Heat transfer coefficients
        self.htc = np.ones(self.Nspm) * 10
        # Cycling experiment
        self.experiment = pybamm.Experiment(
            [
                f"Discharge at {I_app} A for 300 seconds",
            ],
            period="10 seconds",
        )
        # PyBaMM parameters
        self.parameter_values = pybamm.ParameterValues("Chen2020")
        self.managers = ["casadi", "ray"]

    def test_multiprocessing(self):
        for manager in self.managers:
            output1 = lp.solve(
                netlist=self.netlist.copy(),
                parameter_values=self.parameter_values,
                experiment=self.experiment,
                output_variables=None,
                inputs=None,
                initial_soc=0.5,
                nproc=1,
                manager=manager,
            )
            output2 = lp.solve(
                netlist=self.netlist.copy(),
                parameter_values=self.parameter_values,
                experiment=self.experiment,
                output_variables=None,
                inputs=None,
                initial_soc=0.5,
                nproc=2,
                manager=manager,
            )
            a = output1["Terminal voltage [V]"]
            b = output2["Terminal voltage [V]"]
            self.assertEqual(a.shape, (31, 21))
            self.assertTrue(np.allclose(a, b))

    def test_events(self):
        for manager in self.managers:
            output1 = lp.solve(
                netlist=self.netlist.copy(),
                parameter_values=self.parameter_values,
                experiment=self.experiment,
                output_variables=None,
                inputs=None,
                initial_soc=0.01,
                nproc=1,
                manager=manager,
            )
            output2 = lp.solve(
                netlist=self.netlist.copy(),
                parameter_values=self.parameter_values,
                experiment=self.experiment,
                output_variables=None,
                inputs=None,
                initial_soc=0.01,
                nproc=2,
                manager=manager,
            )
            a = output1["Terminal voltage [V]"]
            b = output2["Terminal voltage [V]"]
            self.assertEqual(a.shape, (7, 21))
            self.assertTrue(np.allclose(a, b))

    def test_voltage_limits(self):
        I_app = 5.0
        netlist = lp.setup_circuit(
            Np=1, Ns=1, Rb=1e-4, Rc=1e-2, Ri=3e-2, V=3.6, I=I_app
        )
        parameter_values = pybamm.ParameterValues("Chen2020")
        # Cycling experiment
        charge_exp = pybamm.Experiment(
            [
                f"Charge at {I_app} A for 10000 seconds",
            ],
            period="100 seconds",
        )
        discharge_exp = pybamm.Experiment(
            [
                f"Discharge at {I_app} A for 10000 seconds",
            ],
            period="100 seconds",
        )
        _ = lp.solve(
            netlist=netlist.copy(),
            parameter_values=parameter_values,
            experiment=charge_exp,
            output_variables=None,
            inputs=None,
            initial_soc=0.5,
            nproc=1,
            manager="casadi",
        )
        _ = lp.solve(
            netlist=netlist.copy(),
            parameter_values=parameter_values,
            experiment=discharge_exp,
            output_variables=None,
            inputs=None,
            initial_soc=0.5,
            nproc=1,
            manager="casadi",
        )
        assert True

    def test_rest_step_first(self):
        I_app = 5.0
        netlist = lp.setup_circuit(
            Np=1, Ns=1, Rb=1e-4, Rc=1e-2, Ri=3e-2, V=3.6, I=I_app
        )
        parameter_values = pybamm.ParameterValues("Chen2020")
        # Cycling experiment
        experiment = pybamm.Experiment(
            [
                "Rest for 1 minutes",
                f"Charge at {I_app} A for 1 minutes",
            ],
            period="10 seconds",
        )
        _ = lp.solve(
            netlist=netlist.copy(),
            parameter_values=parameter_values,
            experiment=experiment,
            output_variables=None,
            inputs=None,
            initial_soc=0.5,
            nproc=1,
            manager="casadi",
        )
        assert True


if __name__ == "__main__":
    unittest.main()



#
# Load a netlist to use in a liionpack simulation and compare results to a
# generated one
#

import liionpack as lp
import numpy as np


Rb = 1e-3
Ri = 1e-2
Rc = 1e-2
Rt = 1e-5
Rt2 = 1e-5
I = -2
V = 4.0

net1 = lp.read_netlist(filepath="4p1s.cir", Rb=Rb, Ri=Ri, Rc=Rc, Rt=Rt, I=I, V=V)
V1, I1, _, _, _ = lp.solve_circuit(net1)

net2 = lp.setup_circuit(Np=4, Ns=1, Rb=Rb, Ri=Ri, Rc=Rc, Rt=Rt2, I=I, V=V)
V2, I2, _, _, _ = lp.solve_circuit(net2)

# Nodal order is different
print("V match: ", np.allclose(np.sort(V1), np.sort(V2)))
print("I match: ", np.allclose(np.sort(I1), np.sort(I2)))

lp.draw_circuit(net2, cpt_size=1.0, dpi=300, node_spacing=2.5)



import pybamm
import liionpack as lp
import numpy as np
import unittest
import os
import pandas as pd


class protocolsTest(unittest.TestCase):
    def test_generate_protocol(self):
        experiment = pybamm.Experiment(
            [
                "Charge at 50 A for 30 minutes",
                "Rest for 15 minutes",
                "Discharge at 50 A for 30 minutes",
                "Rest for 15 minutes",
            ],
            period="10 seconds",
        )
        p, t, s = lp.generate_protocol_from_experiment(experiment)
        self.assertEqual(np.sign(p[0][0]), -1)

        experiment = pybamm.Experiment(
            [
                "Discharge at 50 A for 30 minutes",
                "Rest for 15 minutes",
                "Charge at 50 A for 30 minutes",
                "Rest for 15 minutes",
            ],
            period="10 seconds",
        )
        p, t, s = lp.generate_protocol_from_experiment(experiment)
        self.assertEqual(len(p), 4)
        self.assertEqual(np.sign(p[0][0]), 1)

    def test_generate_protocol_from_drive_cycle(self):
        os.chdir(pybamm.__path__[0] + "/..")
        drive_cycle = pd.read_csv(
            pybamm.DataLoader().get_data("US06.csv"), comment="#", header=None
        ).to_numpy()

        experiment = pybamm.Experiment(
            [pybamm.step.current(drive_cycle)], period="1 second"
        )
        p, t, s = lp.generate_protocol_from_experiment(experiment)
        assert len(p[0]) == 601
        assert np.allclose(np.mean(p), 0.8404807891846922)

    def test_current_exception(self):
        def bad_current():
            experiment = pybamm.Experiment(
                [
                    "Charge at 1 C for 30 seconds",
                ],
                period="10 seconds",
            )
            _ = lp.generate_protocol_from_experiment(experiment)

        with self.assertRaises(ValueError):
            bad_current()

    def test_termination(self):
        experiment = pybamm.Experiment(
            [
                "Discharge at 50 A for 30 minutes or until 3.0 V",
            ],
            period="10 seconds",
        )
        p, t, s = lp.generate_protocol_from_experiment(experiment)
        assert t[0] == 3.0


if __name__ == "__main__":
    unittest.main()



"""
Example of running a simulation with two batteries of different initial SOC.
"""

import liionpack as lp
import pybamm

lp.logger.setLevel("NOTICE")

# Define parameters
Np = 2
Ns = 1
Iapp = 5

# Generate the netlist and output variables
netlist = lp.setup_circuit(Np=Np, Ns=Ns, Rb=1.5e-3, Rc=1e-2, Ri=5e-2, V=4.0, I=Iapp)
output_variables = [
    "X-averaged negative particle surface concentration [mol.m-3]",
    "X-averaged positive particle surface concentration [mol.m-3]",
]

# Cycling experiment
experiment = pybamm.Experiment(
    [
        f"Charge at {Iapp} A for 30 minutes",
        "Rest for 15 minutes",
        f"Discharge at {Iapp} A for 30 minutes",
        "Rest for 30 minutes",
    ],
    period="10 seconds",
)

# PyBaMM parameters
param = pybamm.ParameterValues("Chen2020")

concentrations = [lp.update_init_conc(param, SoC=x, update=False) for x in [0.5, 0.6]]

param.update(
    {
        "Initial concentration in negative electrode [mol.m-3]": "[input]",
        "Initial concentration in positive electrode [mol.m-3]": "[input]",
    }
)

inputs = {
    "Initial concentration in negative electrode [mol.m-3]": [
        r[0] for r in concentrations
    ],
    "Initial concentration in positive electrode [mol.m-3]": [
        r[1] for r in concentrations
    ],
}

# Solve pack
output = lp.solve(
    netlist=netlist,
    parameter_values=param,
    experiment=experiment,
    output_variables=output_variables,
    initial_soc=None,
    inputs=inputs,
)

# Plot results
lp.plot_output(output)
lp.show_plots()



#
# Mixed location of terminals and effect on current distribution
#

import liionpack as lp
import pybamm
import matplotlib.pyplot as plt

plt.close("all")

Np = 7
Ns = 1
combos = [
    ["left", "right", "left-right", "right-left"],
    [[0, 0], [-1, -1], [0, -1], [-1, 0], [3, 3]],
]
for terminals in combos:
    lines = ["-", "--", "-.", ":", ".-"]
    plt.figure()
    for i, t in enumerate(terminals):
        netlist = lp.setup_circuit(Np=Np, Ns=Ns, Rb=1e-4, Rc=1e-2, terminals=t)
        param = pybamm.ParameterValues("Chen2020")
        experiment = pybamm.Experiment(
            [
                f"Discharge at {Np} A for 1 minutes",
            ],
            period="10 seconds",
        )
        # Solve pack
        output = lp.solve(
            netlist=netlist,
            parameter_values=param,
            experiment=experiment,
            output_variables=None,
            initial_soc=0.5,
        )

        plt.plot(range(Np), output["Cell current [A]"][-1, :], lines[i], label=t)

    plt.legend()

lp.show_plots()
